/**
  *  \file client/search.h
  *  \brief Search Function Main Code
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_CLIENT_SEARCH_H
#define PCC_V2_CLIENT_SEARCH_H

#include "int/bytecode.h"
#include "game/objlist.h"
#include "cpluslib/string.h"

/** User search query. Contains a search query as entered by the user. */
struct TUserSearchQuery {
    enum Match {
        MatchName,
        MatchTrue,
        MatchFalse,
        MatchLocation
    };

    enum SearchObject {
        // xref WSearchObjectSelector::drawPart
        SearchShips   = 1,
        SearchPlanets = 2,
        SearchBases   = 4,
        SearchUfos    = 8,
        SearchOthers  = 16,
        SearchMineOnly = 32,

        AllObjects    = SearchShips | SearchPlanets | SearchBases | SearchUfos | SearchOthers
    };

    Match    search_kind;
    uint16_t search_obj;
    string_t query;
};

/** Parsed search query. Contains a parsed search query ready for execution. */
struct TParsedSearchQuery {
    // General:
    enum {
        MatchAny,               /**< "Match anything". Used if the user enters an empty query. */
        MatchName,              /**< "Match Name/Id/Comment". Uses the \c id and \c name fields. */
        MatchLocation,          /**< "Match location". Uses the \c location field. */
        MatchExpr               /**< "Match expression". Used for Expression True and Expression False queries. Uses the \c expr_bco field, which contains an expression that returns true on match. */
    } match;
    uint16_t search_obj;

    // Name/Id/Comment query:
    int id;
    string_t name;

    // Location query:
    GPoint location;

    // Expression query:
    IntBCORef expr_bco;
};

/** Search result status. Contains some flags about search query execution to
    permit generation of proper error messages. */
struct TSearchResultStatus {
    /** true iff we had at least one object. */
    bool had_object;
    /** true iff expression was evaluated at least once. */
    bool did_evaluate;
    /** true iff expression had an error at least once. */
    bool had_error;
    /** If expression had an error, its text. */
    string_t last_error;
    
    TSearchResultStatus()
        : had_object(false), did_evaluate(false), had_error(false)
        { }
};

bool prepareSearchQuery(TParsedSearchQuery& query, const TUserSearchQuery& in);
void executeSearchQuery(const TParsedSearchQuery& query, TSearchResultStatus& status, GObjectList& result);

#endif
