/**
  *  \file client/tiles/shipmovement.cc
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <memory>
#include "client/tiles/shipmovement.h"
#include "game/mission.h"
#include "game/ship.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "ui/uiglobal.h"
#include "game/formula.h"
#include "game/host.h"
#include "ui/widgets/button.h"
#include "client/tiles/shipmission.h"
#include "game/univ.h"
#include "game/config.h"
#include "game/shippredict.h"
#include "util/misc.h"

static int
computeCloakFuel(const GShip& ship, int eta)
{
    if (isMissionCloaking(ship.getMission(), ship.getRealOwner())
        && (ship.canDoSpecial(hf_Cloak) || ship.canDoSpecial(hf_HardenedCloak))
        && !ship.canDoSpecial(hf_AdvancedCloak))
    {
        // Regular cloaking that burns fuel
        // PHost/HOST 3.22.20 formula
        const int cfb = config.CloakFuelBurn(ship.getRealOwner());
        int fuel = ship.getHull().getMass() * cfb / 100;
        if (fuel < cfb) {
            fuel = cfb;
        }
        if (eta != 0) {
            fuel *= eta;
        }
        return fuel;
    } else {
        return 0;
    }
}

static int
computeTurnFuel(const GShip& ship, int eta)
{
    int fuel = (int32_t(config.FuelUsagePerTurnFor100KT(ship.getRealOwner())) * ship.getHull().getMass() + 99) / 100;
    if (eta != 0) {
        fuel *= eta;
    }
    return fuel;
}

/*************************** WShipMovementTile ***************************/

WShipMovementTile::WShipMovementTile(GObjectSelection& selection)
    : WObjectSelectionObserverTile(GfxPoint(315, 7*font_heights[FONT_NORMAL]), selection)
{
    init();
}

void
WShipMovementTile::init()
{
    const int he = 7*font_heights[FONT_NORMAL];
    setLayout(RightJustified, 0, false);
    addControl(GfxRect(17, he - 20, 20, 20), h.add(new UIButton("A", 'a')));
    addControl(GfxRect(17, he - 45, 20, 20), h.add(new UIButton("W", 'w')));

    cf_f10.add(h.add(new UIButton(cm_Ship_FleetMenu, "F10", SDLK_F10)).addCommand());
    addControl(GfxRect( 0, he - 72, 39, 24), cf_f10);
}

void
WShipMovementTile::drawShipMovementTile(UIAbstractTableWriter& tw, bool fixed_only)
{
    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Location:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Waypoint:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Distance:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Warp Factor:"));
    tw.writeFixedCell(UIColor::tc_Static, _("E.T.A.:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Fuel Usage:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Engine Load:"));

    if (fixed_only)
        return;                 // save time by not computing movement info

    GShip* ps = dynamic_cast<GShip*>(getCurrentObject());
    GUniverse* univ = getCurrentUniverse();
    if (!ps || ps->getShipKind() != GShip::CurrentShip || !univ)
        return;

    /* compute stuff */
    GShip& s = *ps;
    std::auto_ptr<GShipTurnPredictor> main, towee;
    if (s.getMission() == GMission::msn_Tow && univ->ty_played_ships.isValidIndex(s.getTowId())) {
        towee.reset(new GShipTurnPredictor(*univ, s.getTowId()));
        main.reset(new GShipTurnPredictor(*univ, s.getId(), *towee));
    } else {
        main.reset(new GShipTurnPredictor(*univ, s.getId()));
    }
    main->computeMovement();

    GChunnelData chd;
    parseChunnelMission(*ps, *univ, chd);
    const bool is_chunnel = chd.target != 0 && s.getWaypointDX() == 0 && s.getWaypointDY() == 0;

    const bool is_hyper = s.isHyperdriving();

    tw.newLColumn();

    /* location */
    tw.writeVariableCell(UIColor::tc_Green, univ->getLocationName(s.getPos(), locs_WW | locs_Verbose));

    /* waypoint */
    if (s.getMission() == GMission::msn_Intercept && univ->ty_any_ships.isValidIndex(s.getInterceptId())) {
        tw.writeVariableCell(UIColor::tc_Green, univ->getShip(s.getInterceptId()).getName(GObject::PlainName));
    } else if (is_chunnel) {
        tw.writeVariableCell(UIColor::tc_Green, format(_("chunnel to %s"), univ->getShip(chd.target).getName(GObject::PlainName)));
    } else {
        tw.writeVariableCell(UIColor::tc_Green, univ->getLocationName(s.getWaypoint(), locs_WW | locs_Verbose | locs_Ships));
    }

    /* distance */
    double dist = distFromDX(s.getWaypointDX(), s.getWaypointDY());
    tw.writeVariableCell(is_chunnel
                         ? (chd.failure & GChunnelData::chf_Distance)
                         ? UIColor::tc_Yellow
                         : UIColor::tc_Green
                         : !is_hyper
                         ? UIColor::tc_Green
                         : dist < 20
                         ? UIColor::tc_Red
                         : !host.isExactHyperjumpDistance2(int32(dist*dist+.1)) /* FIXME: this may lose precision, hence the +.1 */
                         ? UIColor::tc_Yellow
                         : UIColor::tc_Green,
                         format(_("%.2f ly"), dist));

    /* warp */
    if (s.getWarp() == 0 && (s.getWaypointDX() | s.getWaypointDY())) {
        tw.writeVariableCell(UIColor::tc_Red, _("not moving"));
    } else {
        // Borland sucks.
        string_t st;
        if (is_hyper)
            st = _("Hyperdrive");
        else if (s.getWarp() == 0 && !is_chunnel)
            st = _("not moving");
        else
            st = format(_("Warp %d"), s.getWarp());
        tw.writeVariableCell(is_chunnel && s.getWarp() > 0
                             ? UIColor::tc_Yellow
                             : UIColor::tc_Green,
                             st);
    }

    /* e.t.a. */
    {
        string_t eta;
        UISkin::Color c = UIColor::tc_Green;

        if (is_chunnel && chd.failure == 0) {
            eta = _("chunnel");
        } else if ((s.getWaypointDX() == 0 && s.getWaypointDY() == 0) && (s.getWarp() > 0 || s.getMission() != GMission::msn_Intercept)) {
            eta = _("at waypoint");
        } else if (is_hyper) {
            eta = _("unknown"); // FIXME?
        } else if (s.getWarp() == 0) {
            c = UIColor::tc_Red;
            eta = _("not moving");
        } else if (main->isAtTurnLimit()) {
            eta = _("too long");
        } else {
            eta = format(_("%d turn%!1{s%}"), main->getNumTurns());
        }
        tw.writeVariableCell(c, eta);
    }

    /* fuel usage */
    {
        string_t fuel;

        /* Recompute turn/cloak fuel usage anew from eta.
           This is how PCC1 does it.
           - also show one turn usage for stationary ships
           - also show full usage if predictor turned off cloaking */
        int eta        = main->getNumTurns();
        int move_fuel  = main->getMovementFuelUsed();
        int cloak_fuel = computeCloakFuel(s, eta);
        int turn_fuel  = computeTurnFuel(s, eta);
        int have_fuel  = s.getCargo(el_Neutronium);

        if (is_chunnel) {
            move_fuel = 50;
            fuel = format(_("chunnel, %d kt"), numToString(move_fuel));
        } else {
            fuel = format(_("%d kt"), numToString(move_fuel));
        }
        if (cloak_fuel > 0 || turn_fuel > 0) {
            if (turn_fuel == 0) {
                fuel += format(_(" (+%d kt cloak)"), cloak_fuel);
            } else {
                fuel += format(_(" (+%d kt)"), cloak_fuel+turn_fuel);
            }
        }

        UISkin::Color c;
        if (move_fuel > have_fuel || (have_fuel == 0 && eta > 0 && !config.AllowNoFuelMovement())) {
            c = UIColor::tc_Red;
        } else if (main->isAtTurnLimit() || move_fuel+cloak_fuel+turn_fuel > have_fuel) {
            c = UIColor::tc_Yellow;
        } else {
            c = UIColor::tc_Green;
        }
        tw.writeVariableCell(c, fuel);
    }

    /* load */
    int towee_mass = 0;
    if (s.getMission() == GMission::msn_Tow && univ->ty_any_ships.isValidIndex(s.getTowId())) {
        mp16_t m = univ->getShip(s.getTowId()).getMass();
        if (m.isKnown()) {
            towee_mass = m;
        }
    }
    tw.writeVariableCell(UIColor::tc_Green, format(_("%d kt"), numToString(s.getMass() + towee_mass)));
    tw.finish();
}

void
WShipMovementTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    UITableWriter tw(ctx, getInnerSize(5, 20), FONT_NORMAL);
    drawShipMovementTile(tw, false);

    // F10 button
    TFrameColor color;
    GShip* ps = dynamic_cast<GShip*>(getCurrentObject());
    if (!ps || ps->getFleetNumber() == 0) {
        color = cf_None;
    } else if (ps->isFleetLeader()) {
        color = cf_Green;
    } else {
        color = cf_Red;
    }
    cf_f10.setColor(color);
}
