/**
  *  \file game/planet.h
  *  \brief Planet / Starbase Management
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_PLANET_H
#define PCC_V2_GAME_PLANET_H

#include "game/cargo.h"
#include "game/coord.h"
#include "game/obj.h"
#include "game/struct.h"
#include "game/unitscore.h"
#include "io/io.h"             // mn16_t etc.

class GShip;
class GUniverse;
class GMessageInformation;

/// Native races.
enum TNativeRace {
    nr_None,
    nr_Humanoid,
    nr_Bovinoid,
    nr_Reptilian,
    nr_Avian,
    nr_Amorphous,
    nr_Insectoid,
    nr_Amphibian,
    nr_Ghipsoldal,
    nr_Siliconoid
};

/// Shipyard actions. Value of the shipyard_action field.
enum TShipyardAction {
    sya_None,                   ///< No operation.
    sya_Fix,                    ///< Fix a ship.
    sya_Recycle                 ///< Recycle a ship.
};

/// Planetary Structures.
enum TPlanetaryStructure {
    str_Mines,
    str_Factories,
    str_Defense,
    str_BaseDefense
};

/** Industrial Activity Ratings. Order matches that used in UTIL.DAT
    record 6 (TUtil6SensorSweep::activity), and is also used in
    GPlanet::getIndustryLevel(int). Note that older PHost versions
    report a sixth level at 150+ structures in UTILx.DAT, this is
    considered a bug and fixed in 4.1a/3.5a. */
enum TIndustryLevel {
    ind_Minimal,                ///< Minimal activity. PHost: 0-29 structures.
    ind_Light,                  ///< Light activity. PHost: 30-59 structures. Host: 0-99 structures.
    ind_Moderate,               ///< Moderate activity. PHost: 60-89 structures.
    ind_Substantial,            ///< Substantial activity. PHost: 90-119 structures.
    ind_Heavy                   ///< Heavy activity. PHost: 120+ structures. Host: 100+ structures.
};

/** Planet. This stores data of a planet and possibly a starbase. It
    is used for played and scanned planets.

    - We always know name and Id of all planets
    - A planet can exist or not:
      - if it exists (isExistant()), we know its position, but not
        necessarily its other data (!= NoPlanet)
      - if it does not exist, we do not know its position. We may still
        know some other data (from scanning) (== NoPlanet)
    - A planet can have three levels of data:
      - if we're playing it, we have full data (hasFullPlanetData() == CurrentPlanet); we may
        have full data for some other planets as well
      - if we've seen it somehow, we have partial data (hasAnyPlanetData() == KnownPlanet).
      - otherwise, we don't know anything about it == UnknownPlanet, NoPlanet
    - A starbase can have two levels of data:
      - if we're playing the planet, the base is either present or not,
        and if it's present (haveBaseData()), we have full data
      - if we're not playing the planet, we only know whether there is
        a base or not. */
class GPlanet : public GMapObject {
 public:
    enum BaseKind {
        UnknownBase,            ///< We do not know whether there is a base.
        NoBase,                 ///< We know that there is no base.
        ExistingBase,           ///< We know that there is a base.
        KnownBase,              ///< We know that there is a base, and have (partial or full) data.
        CurrentBase             ///< We have a BDATA.DAT entry for this base.
    };

    enum PlanetKind {
        NoPlanet,               ///< The planet does not exist. We do not know its position.
        HiddenPlanet,           ///< We do not know this planet's position, but we still have some data from sensor scans.
        UnknownPlanet,          ///< This planet exists but we do not know anything about it but its position.
        KnownPlanet,            ///< This planet exists and we have some information about it.
        CurrentPlanet           ///< We have a PDATAx.DAT entry for this planet.
    };

    enum {
        FlagOwnShipsInOrbit     = 1,    ///< Own ship(s) in orbit.
        FlagAlliedShipsInOrbit  = 2,    ///< Allied ship(s) in orbit.
        FlagEnemyShipsInOrbit   = 4,    ///< Enemy ship(s) in orbit.
        FlagGuessedAllyInOrbit  = 8,    ///< Guessed allied ship(s) in orbit.
        FlagGuessedEnemyInOrbit = 16    ///< Guessed enemy ship(s) in orbit.
    };

    explicit GPlanet(int id);
    explicit GPlanet(const GPlanet&);
    ~GPlanet();

    // Load and Save:
    void        addPlanetData(const TPlanet& data, GPlayerSet source);
    void        addBaseData(const TStarbase& data, GPlayerSet source);
    void        addHistoryData(const TDbPlanet& data);
    void        addMessageInformation(const GMessageInformation& info);  // planet-hist.cc

    void        setXY(int x, int y);
    void        setName(const string_t& name);
    void        setKnownToNotExist(bool value);
    void        setAutobuild(const TAutobuildSettings& abs);

    void        getPlanetData(TPlanet& dat) const;
    void        getBaseData(TStarbase& bdat) const;
    void        getAutobuild(TAutobuildSettings& abs) const;
    void        getHistoryData(TDbPlanet& data) const;

    void        internalCheck();
    void        combinedCheck2(GUniverse& univ);
    int         getOrbitFlags() const;             // FIXME: take this out of GPlanet?
    void        clearOrbitFlags();                 // FIXME: take this out of GPlanet?
    void        setShipInOrbit(const GShip& s);    // FIXME: take this out of GPlanet?

    // GMapObject interface:
    string_t    getName(Name which) const;
    int         getId() const;
    int         getOwner() const;
    GPoint      getPos() const;

    // Planet status accessors:
    bool        isVisible() const;
    GPlayerSet  getPlanetSource() const;
    PlanetKind  getPlanetKind() const;
    bool        hasAnyPlanetData() const;
    bool        hasFullPlanetData() const;
    int16_t     getHistoryTimestamp(TDbPlanetTimestamp which_one) const;

    // Base status accessors:
    GPlayerSet  getBaseSource() const;
    BaseKind    getBaseKind() const;
    bool        hasBase() const;
    bool        hasFullBaseData() const;

    // Owner accessors:
    void        setOwner(mp16_t owner);
    bool        isOwnerKnown() const;

    // Structure accessors:
    mp16_t      getStructures(TPlanetaryStructure ps) const;
    void        setStructures(TPlanetaryStructure ps, mp16_t count);

    mp16_t      getIndustryLevel() const;
    static int  getIndustryLevel(int mifa);
    void        setIndustryLevel(mp16_t level);

    // Colonist accessors:
    mn16_t      getColonistHappiness() const;
    void        setColonistHappiness(mn16_t happiness);
    mp16_t      getColonistTax() const;
    void        setColonistTax(mp16_t tax);
    mn32_t      getColonists() const;
    void        setColonists(mn32_t clans);

    // Native accessors:
    mp16_t      getNativeGovernment() const;
    void        setNativeGovernment(mp16_t gov);
    mn16_t      getNativeHappiness() const;
    void        setNativeHappiness(mn16_t happiness);
    mp16_t      getNativeRace() const;
    void        setNativeRace(mp16_t race);
    mp16_t      getNativeTax() const;
    void        setNativeTax(mp16_t tax);
    void        setNatives(mp32_t natives);
    mp32_t      getNatives() const;

    bool        isKnownToHaveNatives() const;
    void        setKnownToHaveNatives(bool known);

    // FCode accessors:
    string_t    getFCode() const;
    void        setFCode(const string_t& str);
    bool        isFCodeKnown() const;

    // Starbase building accessors:
    bool        isBuildingBase() const;
    void        setBuildBaseFlag(bool b);

    // Environment accessors:
    mp16_t      getOreDensity(int index) const;
    void        setOreDensity(int index, mp16_t amount);
    mp32_t      getOreGround(int index) const;
    void        setOreGround(int index, mp32_t amount);
    mp16_t      getTemperature() const;
    void        setTemperature(mp16_t temp);

    // Cargo accessors:
    mn32_t      getCargo(GCargoType type) const;
    mp32_t      getCargoRaw(GCargoType type) const;
    void        changeCargoRaw(GCargoType type, int32_t amount);
    void        setCargoRaw(GCargoType type, mp32_t amount);

    // Simple base accessors:
    mp16_t      getBaseDamage() const;
    void        setBaseDamage(mp16_t n);
    mp16_t      getBaseMission() const;
    void        setBaseMission(mp16_t mission);
    mp16_t      getBaseTechLevel(TTechLevel level) const;
    void        setBaseTechLevel(TTechLevel level, mp16_t value);

    // Shipyard accessors:
    mp16_t      getBaseShipyardOrder() const;
    mp16_t      getBaseShipyardId() const;
    void        setBaseShipyardOrder(mp16_t action, mp16_t id);

    // Component storage accessors:
    mp16_t      getBaseEngineStore(int slot) const;
    mp16_t      getBaseBeamStore(int slot) const;
    mp16_t      getBaseHullStoreSlot(int slot) const;
    mp16_t      getBaseLauncherStore(int slot) const;
    mp16_t      getBaseAmmoStore(int slot) const;

    void        setBaseEngineStore(int slot, mp16_t amount);
    void        setBaseBeamStore(int slot, mp16_t amount);
    void        setBaseHullStoreSlot(int slot, mp16_t amount);
    void        setBaseLauncherStore(int slot, mp16_t amount);

    // Build order accessors:
    mp16_t      getBaseBuildHull() const;
    TBuildOrder getBaseBuildOrder() const;
    void        setBaseBuildOrder(const TBuildOrder& order);
    mp16_t      getBaseBuildOrderHullSlot() const;

    int         findShipCloningHere(const GUniverse& trn) const;

    // Build queue accessors:
    mp16_t      getBaseQueuePosition() const;
    void        setBaseQueuePosition(mp16_t pos);
    mp32_t      getBaseQueuePriority() const;
    void        setBaseQueuePriority(mp32_t pri);

    // Auto build accessors:
    int         getAutobuildGoal(TPlanetaryStructure ps) const;
    void        setAutobuildGoal(TPlanetaryStructure ps, int value);
    int         getAutobuildSpeed(TPlanetaryStructure ps) const;
    void        setAutobuildSpeed(TPlanetaryStructure ps, int value);

    // Public members:
    GUnitScores unit_scores;     ///< This planet's unit scores (=experience etc.).
    GCargoArbiter cargo_arbiter; ///< Cargo arbiter.

 private:
    // Copy protection
    GPlanet();
    GPlanet& operator=(const GPlanet&);

    // Internal structures:
    struct MapInfo {
        // Always known:
        int id;
        string_t name;

        // From xyplan.dat, util.dat:
        GPoint position;
        bool known_to_not_exist;

        // Derived:
        BaseKind base_kind : 8;
        PlanetKind planet_kind : 8;
        int8_t orbit_flags;

        TAutobuildSettings autobuild;

        MapInfo(int id);
        ~MapInfo();
    };
    struct BaseInfo;
    struct PlanetInfo;

    // Data:
    PlanetInfo*  planet_info;
    BaseInfo*    base_info;
    MapInfo      map_info;

    // Source flags. These specify which players' .dat files contained
    // the PDATA/BDATA records that make up this planet. They do NOT
    // necessarily mean that we (a) play this unit and (b) know
    // everything about it. We will, however, assume that we know
    // everything about the units we play.
    GPlayerSet planet_source;
    GPlayerSet base_source;

    // Data access:
    PlanetInfo&      createPlanetInfo();
    BaseInfo&        createBaseInfo();
    const TPlanet&   getDisplayedPlanet() const;
    const TStarbase& getDisplayedBase() const;
};

#endif
