/**
  *  \file game/racename.cc
  *  \brief Reading and Writing of Race Name files
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/racename.h"
#include "game/player.h"        // getPlayerIdFromChar
#include "game/struct.h"
#include "io/streams.h"
#include "io/io.h"
#include "util/recode.h"
#include "util/nls.h"
#include "game/gamedirs.h"
#include "util/modinit.h"

namespace {
    /** Definition of the race.nm file. */
    enum {
        RN_FULL_ORIG  = 0,
        RN_SHORT_ORIG = RN_FULL_ORIG  + GRaceNameList::FullNameSize*11,
        RN_ADJ_ORIG   = RN_SHORT_ORIG + GRaceNameList::ShortNameSize*11,
        RN_FILE_SIZE  = RN_ADJ_ORIG   + GRaceNameList::AdjNameSize*11
    };
}

/***************************** GRaceNameList *****************************/

/** Default constructor. */
GRaceNameList::GRaceNameList()
{ }

/** Initialize race name list.
    \param s stream pointing to a 'race.nm' file (race names will be
             loaded from that stream) */
GRaceNameList::GRaceNameList(Stream& s)
{
    load(s);
}

/** Destructor. */
GRaceNameList::~GRaceNameList()
{ }

/** Load race names from specified 'race.nm' file.
    \throws FileFormatException if file too short. */
void
GRaceNameList::load(Stream& s)
{
    char buffer[RN_FILE_SIZE];
    s.readT(buffer, sizeof(buffer));

    for(int i = 1; i <= 11; ++i) {
        full_names[i]  = convertGameToUtf8(getBasicStringN(&buffer[RN_FULL_ORIG  + 30*(i-1)], 30));
        short_names[i] = convertGameToUtf8(getBasicStringN(&buffer[RN_SHORT_ORIG + 20*(i-1)], 20));
        adjectives[i]  = convertGameToUtf8(getBasicStringN(&buffer[RN_ADJ_ORIG   + 12*(i-1)], 12));
    }

    full_names[0]  = _("Nobody");
    short_names[0] = _("Nobody");
    adjectives[0]  = _("unowned");

    full_names[12]  = _("The Alien Marauder Alliance");
    short_names[12] = _("Alien Marauders");
    adjectives[12]  = _("Alien");
}

/** Save race names to specified 'race.nm' file. */
void
GRaceNameList::save(Stream& s) const
{
    char buffer[RN_FILE_SIZE];
    for(int i = 1; i <= 11; ++i) {
        storeBasicStringN(&buffer[RN_FULL_ORIG  + 30*(i-1)], 30, convertUtf8ToGame(full_names[i]));
        storeBasicStringN(&buffer[RN_SHORT_ORIG + 20*(i-1)], 20, convertUtf8ToGame(short_names[i]));
        storeBasicStringN(&buffer[RN_ADJ_ORIG   + 12*(i-1)], 12, convertUtf8ToGame(adjectives[i]));
    }
    s.writeT(buffer, sizeof(buffer));
}

/** Get short race name. Returns the short (20 character) name of
    the specified player. Also handles unowned (0) and alien (12). */
const string_t&
GRaceNameList::getShortName(int player) const
{
    ASSERT(player >= 0 && player <= 12);
    return short_names[player];
}

/** Get full race name. Returns the full (30 character) name of
    the specified player. Also handles unowned (0) and alien (12). */
const string_t&
GRaceNameList::getFullName(int player) const
{
    ASSERT(player >= 0 && player <= 12);
    return full_names[player];
}

/** Get race adjective. Returns the adjective (12 characters) of the
    specified player. Also handles unowned (0) and alien (12). */
const string_t&
GRaceNameList::getAdjName(int player) const
{
    ASSERT(player >= 0 && player <= 12);
    return adjectives[player];
}

/** Set short race name. */
void
GRaceNameList::setShortName(int player, const string_t& s)
{
    ASSERT(player >= 0 && player <= 12);
    short_names[player] = s;
}

/** Set full race name. */
void
GRaceNameList::setFullName(int player, const string_t& s)
{
    ASSERT(player >= 0 && player <= 12);
    full_names[player] = s;
}

/** Set race adjective. */
void
GRaceNameList::setAdjName(int player, const string_t& s)
{
    ASSERT(player >= 0 && player <= 12);
    adjectives[player] = s;
}

/** Expand race names.
    "%N" expands to race N's short name.
    "%-N" expands to race N's adjective.
    "%%" is a percent sign. */
string_t
GRaceNameList::expandNames(const string_t text) const
{
    /* Here's an implementation that doesn't modify /text/: */
    string_t rv;
    string_t::size_type pos = 0, n;

    while ((n = text.find('%', pos)) != text.npos) {
        rv.append(text, pos, n - pos);
        ++n;

        bool adj = false;
        if (n < text.size() && text[n] == '-') {
            adj = true;
            ++n;
        }
        if (n < text.size()) {
            if (text[n] == '%') {
                rv += '%';
                ++n;
            } else {
                int race = getPlayerIdFromChar(text[n]);
                if (race > 0 && race <= NUM_OWNERS) {
                    if (adj)
                        rv += getAdjName(race);
                    else
                        rv += getShortName(race);
                    ++n;
                }
            }
        }
        pos = n;
    }
    rv.append(text, pos, text.size() - pos);
    return rv;
}

/******************************** Globals ********************************/

/** Player's race names. Those are race names as seen by the player. Text on screen
    is generated using those. */
GRaceNameList player_racenames;

/** Host's race names. Those are race names used by host to generate messages. Those
    are used for parsing messages. */
GRaceNameList host_racenames;

/** Load Race Names. Loads the RACE.NM file. */
static void
initRaceNames()
{
    Ptr<Stream> file = spec_file_dir->openFile("race.nm", Stream::C_READ);
    player_racenames.load(*file);
    host_racenames = player_racenames;
}

const ModuleInitializer mod_racenames[] = {
    { OnEnterNoGame | OnEnterDirectory, initRaceNames },
    { 0, 0 },
};
