/**
  *  \file spec.h
  *
  *  Specification files. This file contains the classes which
  *  correspond to static game information (hullspec etc.)
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_SPEC_H
#define PCC_V2_SPEC_H

#include <string>
#include <vector>
#include "game/struct.h"
#include "cpluslib/assert.h"
#include "game/hullfunc.h"
#include "game/exp.h"
#include "game/cost.h"
#include "game/playerset.h"
#include "cpluslib/ptrvector.h"
#include "util/modinit.h"
#include "game/specsn.h"

/* Starship Component */
class GComponent {
 public:
    /** Constructor. */
    GComponent(TShortName kind, int id, int mass, int tech, string_t name, GCost cost);
    virtual ~GComponent();

    int getMass() const;
    int getTech() const;
    const GCost& getCost() const;
    string_t getName() const;
    string_t getShortName() const;
    int getId() const;

    void setName(const string_t& name);
    void setShortName(const string_t& shortName);

 private:
    const TShortName kind;
    const int id;

 protected:
    int mass;
    int tech;
    string_t name;
    string_t shortName;
    GCost cost;
};

/* Weapon */
class GWeapon : public GComponent {
 public:
    /** Constructor. */
    GWeapon(TShortName kind, int id, int mass, int tech, string_t name, GCost cost, int kill, int damage)
        : GComponent(kind, id, mass, tech, name, cost),
          kill(kill),
          damage(damage) { }
    /** Get kill (anti-life) power. Should not be 'AC-mangled'. */
    int getKill() const
        { return kill; };
    /** Get damage (anti-structuer) power. */
    int getDamage() const
        { return damage; }
 protected:
    int kill;
    int damage;
};

/* Beam */
class GBeam : public GWeapon {
 public:
    GBeam(int id, const TBeam& data);
 private:
};

/* Torpedo */
class GTorpLauncher;

class GTorpedo : public GWeapon {
 public:
    GTorpedo(const GTorpLauncher& launcher);
};

/* Torpedo Launcher */
class GTorpLauncher : public GWeapon {
 public:
    GTorpLauncher(int id, const TTorpedo& data);

    /** Get cost of one torpedo in mc. */
    int getTorpCost() const
        { return torpCostMC; }

    /** Get torpedo. */
    GTorpedo getTorpedo() const;

 private:
    int torpCostMC;
};

/* Fighter */
class GFighter : public GWeapon {
 public:
    GFighter(int id);
};

/* Engine */
class GEngine : public GComponent {
 public:
    GEngine(int id, const TEngine& data);
    int getFuelFactor(int warp) const;
    int getMaxEfficientWarp() const;
    void setMaxEfficientWarp(int warp);

 private:
    int max_eff_warp;
    int32_t fuel_factors[9];
};

/* Starship Hull */
class GHull : public GComponent {
    struct SpecialAssignment {
        GHullFunctionData::dev_t device;
        GPlayerSet added;
        GPlayerSet removed;

        SpecialAssignment(GHullFunctionData::dev_t device, GPlayerSet added, GPlayerSet removed)
            : device(device), added(added), removed(removed)
        { }
    };

    typedef std::vector<SpecialAssignment> special_t;

    special_t specials[2];

    SpecialAssignment* getSpecialFunctionPtr(GHullFunctionData::dev_t, bool assign_to_hull);

    void enumerateSet(const special_t& stuff, GHullFunctionList& asgn, GPlayerSet player_restriction,
                      GExpLevelSet level_restriction, GHullFunction::Kind kind) const;

    GPlayerSet checkSet(const special_t& stuff, int basic_function,
                        GExpLevelSet levels, bool use_defaults) const;
 public:
    GHull(int id, const THull& data);
    ~GHull();

    int getInternalPictureNumber() const;
    int getExternalPictureNumber() const;
    void setInternalPictureNumber(int n);
    int getFuel() const;
    int getCrew() const;
    int getNumEngines() const;
    int getCargo() const;
    int getNumBays() const;
    int getMaxLaunchers() const;
    int getMaxBeams() const;

    void changeSpecialFunction(GHullFunctionData::dev_t id, GPlayerSet add, GPlayerSet remove, bool assign_to_hull);

    void enumerateHullFunctions(GHullFunctionList& asgn,
                                GPlayerSet player_restriction,
                                GExpLevelSet level_restriction,
                                bool include_new_ship,
                                bool include_racial_abilities) const;

    void clearSpecialFunctions();
    bool canDoSpecial(int basic_function, int player, bool hull, GExpLevelSet levels) const;
    GPlayerSet getPlayersThatCan(int basic_function, bool hull, GExpLevelSet levels) const;

    static special_t racial_abilities;

    static void clearRacialAbilities();
    static void findRacialAbilities();

 private:
    int external_picture;       // picture number as stored in the data file
    int internal_picture;       // picture number used internally
    int max_fuel;
    int max_crew;
    int num_engines;
    int max_cargo;
    int num_bays;
    int max_launchers;
    int max_beams;
};

extern ptr_vector<GHull> hull_spec;
extern TTruehull truehull_spec;

extern const ModuleInitializer mod_spec[];

GBeam&         getBeam(int bid);
GTorpLauncher& getTorp(int tid);
GEngine&       getEngine(int eid);
GHull&         getHull(int hid);
int            getTruehull(int pid, int slot);
int            getTruehullSlot(int pid, int hid);

/** Get fighter for a player.
    \todo maybe cache the objects somehow? */
inline GFighter
getFighterForPlayer(int pid)
{
    return GFighter(pid);
}

/** Get number of hulls. */
inline int
getNumHulls()
{
    return hull_spec.size();
}

/*************************** GComponent Inlines **************************/

/** Get mass of this component, in kilotons. */
inline int
GComponent::getMass() const
{
    return mass;
}

/** Get tech level required to build this component. */
inline int
GComponent::getTech() const
{
    return tech;
}

/** Get cost of one of this components. */
inline const GCost&
GComponent::getCost() const
{
    return cost;
}

/** Get name of this component. This is the full name. */
inline string_t
GComponent::getName() const
{
    return name;
}

/** Get Id number of this component. */
inline int
GComponent::getId() const
{
    return id;
}

/** Set the name of this component. */
inline void
GComponent::setName(const string_t& name)
{
    this->name = name;
}

/** Set the short name of this component.
    Set it to blank to use the default. */
inline void
GComponent::setShortName(const string_t& shortName)
{
    this->shortName = shortName;
}

/***************************** GHull inlines *****************************/

/** Get internal picture number. This is the picture number used by PCC2 to
    render things. */
inline int
GHull::getInternalPictureNumber() const
{
    return internal_picture;
}

/** Get external picture number. This is the picture number seen by the host,
    and used for parsing VCRs. */
inline int
GHull::getExternalPictureNumber() const
{
    return external_picture;
}

/** Set internal picture number. \see getInternalPictureNumber */
inline void
GHull::setInternalPictureNumber(int n)
{
    internal_picture = n;
}

/** Get size of fuel tank. */
inline int
GHull::getFuel() const
{
    return max_fuel;
}

/** Get size of normal crew on this ship. */
inline int
GHull::getCrew() const
{
    return max_crew;
}

/** Get number of engines on this ship. */
inline int
GHull::getNumEngines() const
{
    return num_engines;
}

/** Get size of cargo room on this ship. */
inline int
GHull::getCargo() const
{
    return max_cargo;
}

/** Get number of bays on this ship. */
inline int
GHull::getNumBays() const
{
    return num_bays;
}

/** Get maximum number of torpedo launchers on this ship. */
inline int
GHull::getMaxLaunchers() const
{
    return max_launchers;
}

/** Get maximum number of beam weapons on this ship. */
inline int
GHull::getMaxBeams() const
{
    return max_beams;
}

#endif
