/**
  *  \file int/lex.cc
  *  \brief Interpreter: Tokenizer
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cctype>
#include "int/lex.h"
#include "cpluslib/types.h"
#include "game/formula.h"
#include "int/interror.h"

/****************************** IntTokenizer *****************************/

/** Constructor.
    \param str String to tokenize */
IntTokenizer::IntTokenizer(const string_t& str)
    : line(str),
      pos(0),
      current_token(tEnd),
      current_string(),
      current_int(0),
      current_float(0)
{
    read();
}

/** Destructor. */
IntTokenizer::~IntTokenizer()
{ }

/** Check for token type, and read next if succeeded.
    If the current token is \c t, reads the next token and returns true.
    Otherwise, keep the current token and return false. */
bool
IntTokenizer::checkAdvance(TokenType t)
{
    if (getCurrentToken() == t) {
        read();
        return true;
    } else {
        return false;
    }
}

/** Check for identifier, and read next token if succeeded.
    If the current token is identifier \c keyword, reads the next token and returns true.
    Otherwise, keep the current token and return false. */
bool
IntTokenizer::checkAdvance(const char* keyword)
{
    if (getCurrentToken() == tIdentifier && getCurrentString() == keyword) {
        read();
        return true;
    } else {
        return false;
    }
}

/** Read a token. Advances \c pos and sets \c current_token to the next token encountered
    within the input line. */
void
IntTokenizer::read()
{
    using std::toupper;

    // Skip whitespace
    const string_t::size_type limit = line.size();
    while (pos < limit && (line[pos] == ' ' || line[pos] == '\r' || line[pos] == '\t' || line[pos] == '\f'))
        ++pos;

    // End reached?
    if (pos >= limit) {
        current_token = tEnd;
        return;
    }

    // Check current character
    switch (line[pos]) {
     case '%':
        // Comment
        current_token = tEnd;
        return;
     case '&':
        current_token = tAmpersand;
        ++pos;
        return;
     case '#':
        current_token = tHash;
        ++pos;
        return;
     case '+':
        current_token = tPlus;
        ++pos;
        return;
     case '-':
        ++pos;
        if (pos < limit && line[pos] == '>') {
            current_token = tArrow;
            ++pos;
        } else {
            current_token = tMinus;
        }
        return;
     case '*':
        current_token = tMultiply;
        ++pos;
        return;
     case '/':
        current_token = tSlash;
        ++pos;
        return;
     case '\\':
        current_token = tBackslash;
        ++pos;
        return;
     case '^':
        current_token = tCaret;
        ++pos;
        return;
     case '(':
        current_token = tLParen;
        ++pos;
        return;
     case ')':
        current_token = tRParen;
        ++pos;
        return;
     case ',':
        current_token = tComma;
        ++pos;
        return;
     case '=':
        current_token = tEQ;
        ++pos;
        return;
     case '<':
        ++pos;
        if (pos < limit && line[pos] == '=') {
            current_token = tLE;
            ++pos;
        } else if (pos < limit && line[pos] == '>') {
            current_token = tNE;
            ++pos;
        } else {
            current_token = tLT;
        }
        return;
     case '>':
        ++pos;
        if (pos < limit && line[pos] == '=') {
            current_token = tGE;
            ++pos;
        } else {
            current_token = tGT;
        }
        return;
     case ':':
        ++pos;
        if (pos < limit && line[pos] == '=') {
            current_token = tAssign;
            ++pos;
        } else {
            current_token = tColon;
        }
        return;
     case ';':
        current_token = tSemicolon;
        ++pos;
        return;
     case '.':
        if (pos+1 < limit && line[pos+1] >= '0' && line[pos+1] <= '9') {
            readNumber();
        } else {
            current_token = tDot;
            ++pos;
        }
        return;
     case '\'':
        // Simple string
        // FIXME: sanitize against broken UTF-8 on input? Let TextFile do that?
        {
            string_t::size_type p2 = pos+1;
            while (p2 < limit && line[p2] != '\'')
                ++p2;
            if (p2 >= limit)
                throw IntError::expectSymbol("'");
            current_string.assign(line, pos+1, p2-pos-1);
            current_token = tString;
            pos = p2+1;
        }
        return;

     case '"':
        // String with quotes
        // FIXME: sanitize against broken UTF-8 on input? Let TextFile do that?
        {
            ++pos;
            bool quoted = false;
            current_string.clear();
            while (pos < limit && (line[pos] != '"' || quoted)) {
                if (line[pos] == '\\' && !quoted) {
                    quoted = true;
                } else if (line[pos] == 'n' && quoted) {
                    current_string += "\n";
                    quoted = false;
                } else {
                    current_string += line[pos];
                    quoted = false;
                }
                ++pos;
            }
            if (pos >= limit)
                throw IntError::expectSymbol("\"");
            ++pos;
            current_token = tString;
        }
        return;

     case '0': case '1': case '2': case '3': case '4':
     case '5': case '6': case '7': case '8': case '9':
        readNumber();
        return;

     default:
        char c = toupper(uint8_t(line[pos]));
        ++pos;
        current_string.assign(1, c);
        if ((c >= 'A' && c <= 'Z') || c == '$' || c == '_') {
            // Identifier
            while (pos < limit) {
                c = toupper(uint8_t(line[pos]));
                if ((c >= 'A' && c <= 'Z') || c == '$' || c == '_' || (c >= '0' && c <= '9') || c == '.') {
                    current_string += c;
                    ++pos;
                    if (c != '.' && current_string.size() > 255)
                        throw IntError("Identifier too long");
                } else {
                    break;
                }
            }

            // Special case: if identifier "ends" in '.', strip it.
            if (current_string[current_string.size()-1] == '.') {
                current_string.erase(current_string.size()-1);
                --pos;
            }

            // Handle special keywords
            if (current_string == "TRUE") {
                current_int = 1;
                current_token = tBoolean;
            } else if (current_string == "FALSE") {
                current_int = 0;
                current_token = tBoolean;
            } else if (current_string == "PI") {
                current_token = tFloat;
                current_float = PI;
            } else if (current_string == "AND") {
                current_token = tAND;
            } else if (current_string == "OR") {
                current_token = tOR;
            } else if (current_string == "XOR") {
                current_token = tXOR;
            } else if (current_string == "NOT") {
                current_token = tNOT;
            } else if (current_string == "MOD") {
                current_token = tMOD;
            } else {
                // Identifier
                current_token = tIdentifier;
            }
        } else {
            // Invalid
            current_token = tInvalid;
        }
        return;
    }
}

/** Read number. Assumes that \c pos points to a digit or period. Reads the number
    into current_float / current_int, and sets current_token appropriately. */
void
IntTokenizer::readNumber()
{
    // Initialize
    bool isFloat = false;
    current_int = 0;
    current_float = 0;

    // Read places before period
    while (pos < line.size() && line[pos] >= '0' && line[pos] <= '9') {
        int digit = line[pos] - '0';

        // Advance float
        current_float = 10*current_float + digit;

        // Advance int. Maximum int is 0x7FFFFFFF (yes, this means you cannot enter -0x80000000 in CCScript)
        if (!isFloat) {
            if (current_int > (0x7FFFFFFF - digit) / 10) {
                isFloat = true;
            } else {
                current_int = 10*current_int + digit;
            }
        }

        ++pos;
    }

    // Check for period
    if (pos < line.size() && line[pos] == '.') {
        isFloat = true;
        ++pos;

        // Read places after period
        double divide = 1.0;
        while (pos < line.size() && line[pos] >= '0' && line[pos] <= '9') {
            int digit = line[pos] - '0';
            current_float = 10*current_float + digit;
            divide *= 10;
            ++pos;
        }

        current_float /= divide;
    }

    // Output
    if (isFloat)
        current_token = tFloat;
    else
        current_token = tInteger;
}
